using System;
using System.IO;
using System.Drawing;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using Ventuz.Kernel.Remoting;
using Ventuz.Remoting2.Helpers;
using Ventuz.Remoting2;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// This application demonstrates the remote control of Ventuz by .Net Remoting 2!
// This application can control different scenes. It reads all available scenes
// in a Ventuz project and displayes all externalized properties and methods for a 
// loaded Scene.
//
// This is not an offically supported product. 
// It is a kind of proof-of-concept without proper documentation!
//
// Remember always to link to the latest version of the Ventuz.Kernel.Remoting.dll
// and Ventuz.Kernel.Remoting.Tools.dll!!!
// These DLLs can be found in the Ventuz installation folder.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

namespace RemoteClient
{
	public class RC : System.Windows.Forms.Form, IAsyncHandler
	{
        #region internal class Data2

        class Data2 : ICustomTypeDescriptor
        {
            IExternalCollection extColl;
            PropertyDescriptorCollection pdc;
            RemoteScene scene;

            public Data2(RemoteScene scene, IExternalCollection extColl)
            {
                this.scene = scene;
                this.extColl = extColl;
                this.pdc = new PropertyDescriptorCollection(null);

                List<string> p = new List<string>();
                IEnumerator<IExternal> extEnum = this.extColl.GetEnumerator();

                while ( extEnum.MoveNext() )
                {
                    IExternal ext = extEnum.Current;

                    if ( ext.MajorType == ExternalType.Method )
                        continue;

                    if ( ext.MajorType == ExternalType.Event )
                        continue;

                    ExternalType dataType = ext.MajorType;
                    if ( (dataType & ExternalType.Property) == 0 )
                        continue;

                    if ( ext is IExternalProperty )
                    {
                        this.pdc.Add(new PD((IExternalProperty)ext));
                        p.Add(ext.Name);
                    }
                }

                if ( p.Count > 0 )
                {
                    object[] v = scene.GetProperties(p.ToArray());

                    for ( int i = 0; v != null && i < v.Length; i++ )
                    {
                        ((PD)pdc[i]).SetValueInternal(v[i]);
                    }
                }
            }

            public void Send(ulong when)
            {
                try
                {
                    string[] names = new string[this.pdc.Count];
                    object[] values = new object[this.pdc.Count];

                    for ( int i = 0; i < this.pdc.Count; i++ )
                    {
                        names[i] = pdc[i].Name;
                        values[i] = this.pdc[i].GetValue(null);
                    }

                    if ( ulong.Equals(when, ulong.MaxValue) )
                        scene.SetPropertiesInvokeMethods(names, values, null, null);
                    else
                        scene.SetPropertiesInvokeMethods(names, values, null, null, when, null);

                }
                catch ( Exception ex )
                {
                    Console.WriteLine("Setting Properties/ Invoking methods failed: {0}", ex.Message);
                }
            }

            #region internal class PD

            class PD : PropertyDescriptor
            {
                Type type;
                IExternalProperty ext;
                object value;

                public void SetValueInternal(object value)
                {
                    this.value = value;
                }

                public PD(IExternalProperty ext)
                    : base(ext.Name, null)
                {
                    this.ext = ext;
                    this.type = GetType(ext.MinorType);

                    if ( this.type == typeof(double) )
                        this.type = typeof(float);

                    try
                    {
                        if ( this.type == typeof(string) )
                            this.value = Activator.CreateInstance(type, new char[] { });
                        else
                            this.value = Activator.CreateInstance(type);
                    }
                    catch ( Exception ex )
                    {
                        Console.WriteLine("PD Error: {0}", ex.Message);
                    }
                }

                private Type GetType(ExternalType extType)
                {
                    switch ( extType & Ventuz.Remoting2.ExternalType.MinorTypeMask )
                    {
                        case ExternalType.AssetURI:
                            return typeof(string);
                        case ExternalType.Boolean:
                            return typeof(bool);
                        case ExternalType.Color:
                            return typeof(System.Drawing.Color);
                        case ExternalType.Enumeration:
                            return typeof(string);
                        case ExternalType.Float:
                            return typeof(float);
                        case ExternalType.Integer:
                            return typeof(int);
                        case ExternalType.String:
                            return typeof(string);
                    }

                    return typeof(string);
                }

                #region Overrides

                public override bool CanResetValue(object component)
                {
                    return false;
                }

                public override Type ComponentType
                {
                    get { return typeof(object); }
                }

                public override object GetValue(object component)
                {
                    return value;
                }

                public override void SetValue(object component, object value)
                {
                    this.value = value;
                }

                public override bool IsReadOnly
                {
                    get { return false; }
                }

                public override Type PropertyType
                {
                    get { return type; }
                }

                public override void ResetValue(object component)
                {
                }

                public override bool ShouldSerializeValue(object component)
                {
                    return false;
                }

                #endregion
            }

            #endregion

            #region ICustomTypeDescriptor Members

            public TypeConverter GetConverter()
            {
                return TypeDescriptor.GetConverter(typeof(Data2), true);
            }

            public EventDescriptorCollection GetEvents(Attribute[] attributes)
            {
                return new EventDescriptorCollection(new EventDescriptor[0]);
            }

            EventDescriptorCollection System.ComponentModel.ICustomTypeDescriptor.GetEvents()
            {
                return new EventDescriptorCollection(new EventDescriptor[0]);
            }

            public string GetComponentName()
            {
                return "Data";
            }

            public object GetPropertyOwner(PropertyDescriptor pd)
            {
                return this;
            }

            AttributeCollection ac = new AttributeCollection(new Attribute[0]);

            public AttributeCollection GetAttributes()
            {
                return ac;
            }

            public PropertyDescriptorCollection GetProperties(Attribute[] attributes)
            {
                return this.pdc;
            }

            PropertyDescriptorCollection System.ComponentModel.ICustomTypeDescriptor.GetProperties()
            {
                return GetProperties(null);
            }

            public object GetEditor(Type editorBaseType)
            {
                return TypeDescriptor.GetEditor(this, editorBaseType, true);
            }

            public PropertyDescriptor GetDefaultProperty()
            {
                return null;
            }

            public EventDescriptor GetDefaultEvent()
            {
                return null;
            }

            public string GetClassName()
            {
                return TypeDescriptor.GetClassName(this, true);
            }

            #endregion

        } // Data2

        #endregion


		#region Designer members
		private System.Windows.Forms.Button btConnect;
		private System.Windows.Forms.Button btRender;
		private System.Windows.Forms.PropertyGrid propertyGrid1;
		private System.Windows.Forms.Button btUnloadall;
		private System.Windows.Forms.Button btSend;
		private System.Windows.Forms.Panel PanelRemoteMethods;
		private System.Windows.Forms.Panel panel_SceneBrowse;
		private System.Windows.Forms.Panel panel_SB_Buttons;
		private System.Windows.Forms.Panel panelRemoteProps;
		private System.Windows.Forms.TextBox tb_Host;
		private System.Windows.Forms.TreeView tree;
		private System.Windows.Forms.ImageList imageList1;
		private System.Timers.Timer timer1;
		private System.Windows.Forms.Button btUnload;
		private System.Windows.Forms.Panel panel_Log;
		private System.Windows.Forms.TextBox tbEvents;
		private System.Windows.Forms.Button btSnapshot;
        private CheckBox cb_AutoUpdate;
        private SplitContainer splitContainer1;
        private Button bt_R2D;
        private Button bt_MachineID;
        private Label lblVersion;
        private Button bt_Out_In;
        private Button btRefreshExt;
        private Button button1;
		private System.ComponentModel.IContainer components;
		#endregion

		#region Constructor

		public RC()
		{
			InitializeComponent();
		}

		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#endregion

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
            this.components = new System.ComponentModel.Container();
            System.ComponentModel.ComponentResourceManager resources = new System.ComponentModel.ComponentResourceManager(typeof(RC));
            this.btConnect = new System.Windows.Forms.Button();
            this.btRender = new System.Windows.Forms.Button();
            this.propertyGrid1 = new System.Windows.Forms.PropertyGrid();
            this.btUnloadall = new System.Windows.Forms.Button();
            this.btSend = new System.Windows.Forms.Button();
            this.PanelRemoteMethods = new System.Windows.Forms.Panel();
            this.panel_SceneBrowse = new System.Windows.Forms.Panel();
            this.tree = new System.Windows.Forms.TreeView();
            this.imageList1 = new System.Windows.Forms.ImageList(this.components);
            this.tb_Host = new System.Windows.Forms.TextBox();
            this.panel_SB_Buttons = new System.Windows.Forms.Panel();
            this.bt_Out_In = new System.Windows.Forms.Button();
            this.lblVersion = new System.Windows.Forms.Label();
            this.bt_MachineID = new System.Windows.Forms.Button();
            this.bt_R2D = new System.Windows.Forms.Button();
            this.cb_AutoUpdate = new System.Windows.Forms.CheckBox();
            this.btUnload = new System.Windows.Forms.Button();
            this.panelRemoteProps = new System.Windows.Forms.Panel();
            this.btRefreshExt = new System.Windows.Forms.Button();
            this.btSnapshot = new System.Windows.Forms.Button();
            this.timer1 = new System.Timers.Timer();
            this.panel_Log = new System.Windows.Forms.Panel();
            this.tbEvents = new System.Windows.Forms.TextBox();
            this.splitContainer1 = new System.Windows.Forms.SplitContainer();
            this.button1 = new System.Windows.Forms.Button();
            this.panel_SceneBrowse.SuspendLayout();
            this.panel_SB_Buttons.SuspendLayout();
            this.panelRemoteProps.SuspendLayout();
            ((System.ComponentModel.ISupportInitialize)(this.timer1)).BeginInit();
            this.panel_Log.SuspendLayout();
            ((System.ComponentModel.ISupportInitialize)(this.splitContainer1)).BeginInit();
            this.splitContainer1.Panel1.SuspendLayout();
            this.splitContainer1.Panel2.SuspendLayout();
            this.splitContainer1.SuspendLayout();
            this.SuspendLayout();
            // 
            // btConnect
            // 
            this.btConnect.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.btConnect.Location = new System.Drawing.Point(3, 8);
            this.btConnect.Name = "btConnect";
            this.btConnect.Size = new System.Drawing.Size(75, 23);
            this.btConnect.TabIndex = 0;
            this.btConnect.Text = "Connect";
            this.btConnect.Click += new System.EventHandler(this.btConnect_Click);
            // 
            // btRender
            // 
            this.btRender.Enabled = false;
            this.btRender.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.btRender.Location = new System.Drawing.Point(3, 37);
            this.btRender.Name = "btRender";
            this.btRender.Size = new System.Drawing.Size(75, 23);
            this.btRender.TabIndex = 2;
            this.btRender.Text = "Render";
            this.btRender.Click += new System.EventHandler(this.btRender_Click);
            // 
            // propertyGrid1
            // 
            this.propertyGrid1.BackColor = System.Drawing.SystemColors.ScrollBar;
            this.propertyGrid1.CategoryForeColor = System.Drawing.SystemColors.InactiveCaptionText;
            this.propertyGrid1.Cursor = System.Windows.Forms.Cursors.HSplit;
            this.propertyGrid1.Dock = System.Windows.Forms.DockStyle.Fill;
            this.propertyGrid1.HelpVisible = false;
            this.propertyGrid1.LineColor = System.Drawing.SystemColors.ScrollBar;
            this.propertyGrid1.Location = new System.Drawing.Point(5, 28);
            this.propertyGrid1.Name = "propertyGrid1";
            this.propertyGrid1.Size = new System.Drawing.Size(140, 316);
            this.propertyGrid1.TabIndex = 3;
            this.propertyGrid1.ToolbarVisible = false;
            this.propertyGrid1.ViewBackColor = System.Drawing.SystemColors.ScrollBar;
            this.propertyGrid1.SelectedObjectsChanged += new System.EventHandler(this.propertyGrid1_SelectedObjectsChanged);
            // 
            // btUnloadall
            // 
            this.btUnloadall.Enabled = false;
            this.btUnloadall.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.btUnloadall.Location = new System.Drawing.Point(84, 66);
            this.btUnloadall.Name = "btUnloadall";
            this.btUnloadall.Size = new System.Drawing.Size(75, 23);
            this.btUnloadall.TabIndex = 2;
            this.btUnloadall.Text = "Unload all";
            this.btUnloadall.Click += new System.EventHandler(this.btUnloadall_Click);
            // 
            // btSend
            // 
            this.btSend.Dock = System.Windows.Forms.DockStyle.Bottom;
            this.btSend.Enabled = false;
            this.btSend.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.btSend.Location = new System.Drawing.Point(5, 344);
            this.btSend.Name = "btSend";
            this.btSend.Size = new System.Drawing.Size(140, 36);
            this.btSend.TabIndex = 4;
            this.btSend.Text = "Send Data";
            this.btSend.Click += new System.EventHandler(this.btSend_Click);
            // 
            // PanelRemoteMethods
            // 
            this.PanelRemoteMethods.AutoScroll = true;
            this.PanelRemoteMethods.Dock = System.Windows.Forms.DockStyle.Fill;
            this.PanelRemoteMethods.Location = new System.Drawing.Point(0, 0);
            this.PanelRemoteMethods.Name = "PanelRemoteMethods";
            this.PanelRemoteMethods.Size = new System.Drawing.Size(190, 421);
            this.PanelRemoteMethods.TabIndex = 5;
            // 
            // panel_SceneBrowse
            // 
            this.panel_SceneBrowse.Controls.Add(this.tree);
            this.panel_SceneBrowse.Controls.Add(this.tb_Host);
            this.panel_SceneBrowse.Controls.Add(this.panel_SB_Buttons);
            this.panel_SceneBrowse.Dock = System.Windows.Forms.DockStyle.Left;
            this.panel_SceneBrowse.Location = new System.Drawing.Point(0, 0);
            this.panel_SceneBrowse.Name = "panel_SceneBrowse";
            this.panel_SceneBrowse.Padding = new System.Windows.Forms.Padding(5);
            this.panel_SceneBrowse.Size = new System.Drawing.Size(344, 421);
            this.panel_SceneBrowse.TabIndex = 6;
            // 
            // tree
            // 
            this.tree.BackColor = System.Drawing.SystemColors.ScrollBar;
            this.tree.Dock = System.Windows.Forms.DockStyle.Fill;
            this.tree.ImageIndex = 0;
            this.tree.ImageList = this.imageList1;
            this.tree.Location = new System.Drawing.Point(5, 25);
            this.tree.Name = "tree";
            this.tree.SelectedImageIndex = 0;
            this.tree.Size = new System.Drawing.Size(334, 296);
            this.tree.TabIndex = 8;
            this.tree.AfterSelect += new System.Windows.Forms.TreeViewEventHandler(this.tree_AfterSelect);
            this.tree.DoubleClick += new System.EventHandler(this.tree_DoubleClick);
            // 
            // imageList1
            // 
            this.imageList1.ImageStream = ((System.Windows.Forms.ImageListStreamer)(resources.GetObject("imageList1.ImageStream")));
            this.imageList1.TransparentColor = System.Drawing.Color.Transparent;
            this.imageList1.Images.SetKeyName(0, "");
            this.imageList1.Images.SetKeyName(1, "");
            this.imageList1.Images.SetKeyName(2, "");
            this.imageList1.Images.SetKeyName(3, "");
            // 
            // tb_Host
            // 
            this.tb_Host.Dock = System.Windows.Forms.DockStyle.Top;
            this.tb_Host.Location = new System.Drawing.Point(5, 5);
            this.tb_Host.Name = "tb_Host";
            this.tb_Host.Size = new System.Drawing.Size(334, 20);
            this.tb_Host.TabIndex = 4;
            this.tb_Host.Text = "127.0.0.1";
            // 
            // panel_SB_Buttons
            // 
            this.panel_SB_Buttons.Controls.Add(this.bt_Out_In);
            this.panel_SB_Buttons.Controls.Add(this.lblVersion);
            this.panel_SB_Buttons.Controls.Add(this.bt_MachineID);
            this.panel_SB_Buttons.Controls.Add(this.bt_R2D);
            this.panel_SB_Buttons.Controls.Add(this.cb_AutoUpdate);
            this.panel_SB_Buttons.Controls.Add(this.btRender);
            this.panel_SB_Buttons.Controls.Add(this.btUnloadall);
            this.panel_SB_Buttons.Controls.Add(this.button1);
            this.panel_SB_Buttons.Controls.Add(this.btConnect);
            this.panel_SB_Buttons.Controls.Add(this.btUnload);
            this.panel_SB_Buttons.Dock = System.Windows.Forms.DockStyle.Bottom;
            this.panel_SB_Buttons.Location = new System.Drawing.Point(5, 321);
            this.panel_SB_Buttons.Name = "panel_SB_Buttons";
            this.panel_SB_Buttons.Size = new System.Drawing.Size(334, 95);
            this.panel_SB_Buttons.TabIndex = 3;
            // 
            // bt_Out_In
            // 
            this.bt_Out_In.Location = new System.Drawing.Point(85, 37);
            this.bt_Out_In.Name = "bt_Out_In";
            this.bt_Out_In.Size = new System.Drawing.Size(74, 23);
            this.bt_Out_In.TabIndex = 7;
            this.bt_Out_In.Text = "Prepare";
            this.bt_Out_In.UseVisualStyleBackColor = true;
            this.bt_Out_In.Click += new System.EventHandler(this.bt_Prepare_Click);
            // 
            // lblVersion
            // 
            this.lblVersion.AutoSize = true;
            this.lblVersion.Location = new System.Drawing.Point(88, 13);
            this.lblVersion.Name = "lblVersion";
            this.lblVersion.Size = new System.Drawing.Size(0, 13);
            this.lblVersion.TabIndex = 6;
            // 
            // bt_MachineID
            // 
            this.bt_MachineID.Location = new System.Drawing.Point(173, 66);
            this.bt_MachineID.Name = "bt_MachineID";
            this.bt_MachineID.Size = new System.Drawing.Size(155, 23);
            this.bt_MachineID.TabIndex = 5;
            this.bt_MachineID.Text = "Machine Signature";
            this.bt_MachineID.UseVisualStyleBackColor = true;
            this.bt_MachineID.Click += new System.EventHandler(this.bt_MachineID_Click);
            // 
            // bt_R2D
            // 
            this.bt_R2D.Location = new System.Drawing.Point(173, 37);
            this.bt_R2D.Name = "bt_R2D";
            this.bt_R2D.Size = new System.Drawing.Size(155, 23);
            this.bt_R2D.TabIndex = 4;
            this.bt_R2D.Text = "Render to disk";
            this.bt_R2D.UseVisualStyleBackColor = true;
            this.bt_R2D.Click += new System.EventHandler(this.bt_R2D_Click);
            // 
            // cb_AutoUpdate
            // 
            this.cb_AutoUpdate.AutoSize = true;
            this.cb_AutoUpdate.Location = new System.Drawing.Point(181, 8);
            this.cb_AutoUpdate.Name = "cb_AutoUpdate";
            this.cb_AutoUpdate.Size = new System.Drawing.Size(140, 17);
            this.cb_AutoUpdate.TabIndex = 3;
            this.cb_AutoUpdate.Text = "Autoupdate Scene-Tree";
            this.cb_AutoUpdate.UseVisualStyleBackColor = true;
            // 
            // btUnload
            // 
            this.btUnload.Enabled = false;
            this.btUnload.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.btUnload.Location = new System.Drawing.Point(3, 66);
            this.btUnload.Name = "btUnload";
            this.btUnload.Size = new System.Drawing.Size(75, 23);
            this.btUnload.TabIndex = 2;
            this.btUnload.Text = "Unload";
            this.btUnload.Click += new System.EventHandler(this.btUnload_Click);
            // 
            // panelRemoteProps
            // 
            this.panelRemoteProps.Controls.Add(this.propertyGrid1);
            this.panelRemoteProps.Controls.Add(this.btRefreshExt);
            this.panelRemoteProps.Controls.Add(this.btSend);
            this.panelRemoteProps.Controls.Add(this.btSnapshot);
            this.panelRemoteProps.Dock = System.Windows.Forms.DockStyle.Fill;
            this.panelRemoteProps.Location = new System.Drawing.Point(0, 0);
            this.panelRemoteProps.Name = "panelRemoteProps";
            this.panelRemoteProps.Padding = new System.Windows.Forms.Padding(5);
            this.panelRemoteProps.Size = new System.Drawing.Size(150, 421);
            this.panelRemoteProps.TabIndex = 7;
            // 
            // btRefreshExt
            // 
            this.btRefreshExt.AutoSizeMode = System.Windows.Forms.AutoSizeMode.GrowAndShrink;
            this.btRefreshExt.Dock = System.Windows.Forms.DockStyle.Top;
            this.btRefreshExt.Location = new System.Drawing.Point(5, 5);
            this.btRefreshExt.Name = "btRefreshExt";
            this.btRefreshExt.Size = new System.Drawing.Size(140, 23);
            this.btRefreshExt.TabIndex = 6;
            this.btRefreshExt.Text = "Refresh Externals";
            this.btRefreshExt.UseVisualStyleBackColor = true;
            this.btRefreshExt.Click += new System.EventHandler(this.btRefreshExt_Click);
            // 
            // btSnapshot
            // 
            this.btSnapshot.Dock = System.Windows.Forms.DockStyle.Bottom;
            this.btSnapshot.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.btSnapshot.Location = new System.Drawing.Point(5, 380);
            this.btSnapshot.Name = "btSnapshot";
            this.btSnapshot.Size = new System.Drawing.Size(140, 36);
            this.btSnapshot.TabIndex = 5;
            this.btSnapshot.Text = "Snapshot";
            this.btSnapshot.Click += new System.EventHandler(this.btSnapshot_Click);
            // 
            // timer1
            // 
            this.timer1.Enabled = true;
            this.timer1.Interval = 2000D;
            this.timer1.SynchronizingObject = this;
            this.timer1.Elapsed += new System.Timers.ElapsedEventHandler(this.timer1_Elapsed);
            // 
            // panel_Log
            // 
            this.panel_Log.Controls.Add(this.tbEvents);
            this.panel_Log.Dock = System.Windows.Forms.DockStyle.Bottom;
            this.panel_Log.Location = new System.Drawing.Point(0, 421);
            this.panel_Log.Name = "panel_Log";
            this.panel_Log.Size = new System.Drawing.Size(688, 72);
            this.panel_Log.TabIndex = 8;
            // 
            // tbEvents
            // 
            this.tbEvents.BackColor = System.Drawing.SystemColors.ControlDark;
            this.tbEvents.Dock = System.Windows.Forms.DockStyle.Fill;
            this.tbEvents.Location = new System.Drawing.Point(0, 0);
            this.tbEvents.Multiline = true;
            this.tbEvents.Name = "tbEvents";
            this.tbEvents.ReadOnly = true;
            this.tbEvents.ScrollBars = System.Windows.Forms.ScrollBars.Vertical;
            this.tbEvents.Size = new System.Drawing.Size(688, 72);
            this.tbEvents.TabIndex = 0;
            // 
            // splitContainer1
            // 
            this.splitContainer1.Dock = System.Windows.Forms.DockStyle.Fill;
            this.splitContainer1.Location = new System.Drawing.Point(344, 0);
            this.splitContainer1.Name = "splitContainer1";
            // 
            // splitContainer1.Panel1
            // 
            this.splitContainer1.Panel1.Controls.Add(this.panelRemoteProps);
            this.splitContainer1.Panel1MinSize = 150;
            // 
            // splitContainer1.Panel2
            // 
            this.splitContainer1.Panel2.Controls.Add(this.PanelRemoteMethods);
            this.splitContainer1.Panel2MinSize = 100;
            this.splitContainer1.Size = new System.Drawing.Size(344, 421);
            this.splitContainer1.SplitterDistance = 150;
            this.splitContainer1.TabIndex = 0;
            // 
            // button1
            // 
            this.button1.FlatStyle = System.Windows.Forms.FlatStyle.System;
            this.button1.Location = new System.Drawing.Point(85, 8);
            this.button1.Name = "button1";
            this.button1.Size = new System.Drawing.Size(75, 23);
            this.button1.TabIndex = 0;
            this.button1.Text = "AutoConnect";
            this.button1.Click += new System.EventHandler(this.btAutoConnect_Click);
            // 
            // RC
            // 
            this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
            this.BackColor = System.Drawing.SystemColors.ButtonShadow;
            this.ClientSize = new System.Drawing.Size(688, 493);
            this.Controls.Add(this.splitContainer1);
            this.Controls.Add(this.panel_SceneBrowse);
            this.Controls.Add(this.panel_Log);
            this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
            this.Name = "RC";
            this.Text = "Remote Control V4 (Remoting 2)";
            this.panel_SceneBrowse.ResumeLayout(false);
            this.panel_SceneBrowse.PerformLayout();
            this.panel_SB_Buttons.ResumeLayout(false);
            this.panel_SB_Buttons.PerformLayout();
            this.panelRemoteProps.ResumeLayout(false);
            ((System.ComponentModel.ISupportInitialize)(this.timer1)).EndInit();
            this.panel_Log.ResumeLayout(false);
            this.panel_Log.PerformLayout();
            this.splitContainer1.Panel1.ResumeLayout(false);
            this.splitContainer1.Panel2.ResumeLayout(false);
            ((System.ComponentModel.ISupportInitialize)(this.splitContainer1)).EndInit();
            this.splitContainer1.ResumeLayout(false);
            this.ResumeLayout(false);

		}
		#endregion


        #region static MAIN
        
        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main()
        {
            // TODO: why
            try
            {
                throw new Exception();
            }
            catch { }

            Application.EnableVisualStyles();
            Application.DoEvents();
            Application.Run(new RC());
        } 

        #endregion


        //-------------------------------------
        //          Remoting 2
        RemoteScene ventuz2 = null;                     // the active layout scene
        RemoteScene scene2 = null;                      // manualy loaded scene
        //
        //-------------------------------------

        Data2 data;
        Hashtable name2nodes = new Hashtable();

        int evCount = 0;                                // event counter
        Queue<string> eventQ = new Queue<string>();     // event message queue
        bool haveNewEvents = false;
        delegate void UpdateEventText();


        #region G U I   E V E N T S

        #region CONNECT

        /// <summary>
        /// -- CONNECT --
        /// Connect to Ventuz server and build scene tree
        /// </summary>
        private void btConnect_Click(object sender, System.EventArgs e)
        {
            try
            {
                tree.BeginUpdate();

                tree.Nodes.Clear();
                name2nodes.Clear();

                // clear event list
                tbEvents.Text = "";

                // connect to Ventuz
                this.ventuz2 = RemoteScene.Connect(tb_Host.Text);
                ulong test = this.ventuz2.Now;                      // test if we are connected to a Ventuz server because Connect does not throw an Exception if no server available.

                //btConnect.Enabled = false;
                btConnect.Text = "Re-Connect";
                lblVersion.Text = string.Format("Vers.: {0}", this.ventuz2.VentuzInfo.Version.ToString(4));
                btUnloadall.Enabled = true;
                btUnload.Enabled = true;

                // get and display available scenes
                string[] names = ventuz2.GetAvailableScenes();
                RemoteScene[] loaded = ventuz2.GetLoadedScenes();

                if ( names != null )
                {
                    foreach ( string name in names )
                    {
                        if ( name == null )
                            continue;

                        // Remoting 2: scenes come without leading'/'
                        string newName = name;
                        if ( !newName.StartsWith("/") )
                            newName = "/" + name;

                        string[] tok = newName.Split('/');
                        if ( tok.Length < 1 )
                            continue;

                        TreeNode node = GetNode(tree.Nodes, tok, 1, newName);
                    }
                }

                tree.EndUpdate();
            }
            catch ( Exception ex )
            {
                MessageBox.Show(this, ex.Message);
            }
        }

        private void btAutoConnect_Click(object sender, System.EventArgs e)
        {
            bool succeeded = false;
            int MAX = 1000;

            while ( ! succeeded && MAX > 0 )
            {
                try
                {
                    tree.BeginUpdate();

                    tree.Nodes.Clear();
                    name2nodes.Clear();

                    // clear event list
                    tbEvents.Text = "";

                    // connect to Ventuz
                    this.ventuz2 = RemoteScene.Connect(tb_Host.Text);
                    ulong test = this.ventuz2.Now;                      // test if we are connected to a Ventuz server because Connect does not throw an Exception if no server available.

                    //btConnect.Enabled = false;
                    btConnect.Text = "Re-Connect";
                    lblVersion.Text = string.Format("Vers.: {0}", this.ventuz2.VentuzInfo.Version.ToString(4));
                    btUnloadall.Enabled = true;
                    btUnload.Enabled = true;

                    // get and display available scenes
                    string[] names = ventuz2.GetAvailableScenes();
                    RemoteScene[] loaded = ventuz2.GetLoadedScenes();

                    if ( names != null )
                    {
                        foreach ( string name in names )
                        {
                            if ( name == null )
                                continue;

                            // Remoting 2: scenes come without leading'/'
                            string newName = name;
                            if ( !newName.StartsWith("/") )
                                newName = "/" + name;

                            string[] tok = newName.Split('/');
                            if ( tok.Length < 1 )
                                continue;

                            TreeNode node = GetNode(tree.Nodes, tok, 1, newName);
                        }
                    }

                    tree.EndUpdate();

                    succeeded = true;
                }
                catch ( Exception ex )
                {
                    Console.WriteLine("CONNECT Error: {0}", ex.Message);
                    MAX--;
                } 
            }
        }

        #endregion

        #region RENDER

        /// <summary>
        /// -- RENDER --
        /// Load and activate selected Ventuz scene.
        /// </summary>
        private void btRender_Click(object sender, System.EventArgs e)
        {
            if ( this.ventuz2 != null && tree.SelectedNode != null && tree.SelectedNode.Tag != null )
            {
                try
                {
                    if ( this.scene2 != null )
                    {
                        // TODO: unsubscribe all scene events ???
                        try
                        {
                            this.scene2.Callback -= new SceneCallbackHandler(Scene_Callback2);
                        }
                        catch { }   // scene2 could have been already disposed on server side
                        this.scene2 = null;
                    }

                    propertyGrid1.SelectedObject = null;

                    // load scene
                    string sceneName = (string)tree.SelectedNode.Tag;
                    this.scene2 = this.ventuz2.LoadScene(sceneName, true);

                    // pre-validate (Can be done, but is not necessary. Otherwise a validation is processed if the scene is set on a port.)
                    this.scene2.Validate();

                    // set to port and activate
                    this.ventuz2.Ports[0].Scene = this.scene2;
                    this.ventuz2.Ports[0].Active = true;

                    // listen to all scene events?
                    // NOTE: This is necessary even if you want to receive single events (Remoting DLL version 3.03.00)!
                    // (see SetupMethodsAndEvents())
                    this.scene2.Callback += new SceneCallbackHandler(Scene_Callback2);
                    
                    // get properties and methods
                    RefreshExternals(this.scene2);

                    timer1_Elapsed(this, null);
                    btUnload.Enabled = true;
                    bt_R2D.Enabled = true;

                    tree.SelectedNode.SelectedImageIndex = 2;
                }
                catch ( Exception ex )
                {
                    Console.WriteLine(ex.Message);
                }
            }
        }

        private void tree_DoubleClick(object sender, System.EventArgs e)
        {
            btRender_Click(this, EventArgs.Empty);
        }

        #endregion

        #region UNLOAD

        /// <summary>
        /// UNLOAD active scene
        /// </summary>
        private void btUnload_Click(object sender, System.EventArgs e)
        {
            if ( ventuz2 != null )
            {
                RemoteScene rs = ventuz2.Ports[0].Scene;
                ventuz2.Ports[0].Scene = null;

                if ( rs != null )
                {
                    try
                    {
                        rs.Callback -= new SceneCallbackHandler(Scene_Callback2);
                    }
                    catch { }

                    rs.ReleaseScene();
                }

                if ( tree.SelectedNode != null )
                    tree.SelectedNode.SelectedImageIndex = 1;
                btRender.Enabled = true;
                btUnload.Enabled = false;
                bt_R2D.Enabled = false;

                propertyGrid1.SelectedObject = null;
            }
        }

        /// <summary>
        /// UNLOAD ALL scenes
        /// </summary>
        private void btUnloadall_Click(object sender, System.EventArgs e)
        {
            if ( this.ventuz2 != null )
            {
                // unload all scenes
                RemoteScene[] allScenes = this.ventuz2.GetLoadedScenes();

                Console.WriteLine("Unloading {0} scenes...", allScenes.Length);

                foreach ( RemoteScene rs in allScenes ) 
                {
                    if ( rs.Port != null )
                        //rs.Port.SetActive(false, 0, null);
                        rs.Port = null;
                    rs.ReleaseScene();
                }

                if ( tree.SelectedNode != null )
                    tree.SelectedNode.SelectedImageIndex = 1;
                btRender.Enabled = true;
                btUnload.Enabled = false;
                bt_R2D.Enabled = false;

                propertyGrid1.SelectedObject = null;
            }
        } 

        #endregion

        #region PREPARE

        private void bt_Prepare_Click(object sender, EventArgs e)
        {
            if ( this.ventuz2 != null && tree.SelectedNode != null && tree.SelectedNode.Tag != null )
            {
                try
                {
                    propertyGrid1.SelectedObject = null;

                    // prepare scene by loading and validation
                    // NOTE: unfortunatly this will not create the ExternalsCollection for the scene!
                    string sceneName = (string)tree.SelectedNode.Tag;
                    RemoteScene preparedScene = null;
                    uint cmdID = this.ventuz2.LoadScene(sceneName, true, this, out preparedScene);              // LOAD
                    preparedScene.Validate(RemoteScene.IMMEDIATE_BLOCKING, this);                               // VALIDATE
                    preparedScene.SyncPoint(1111, RemoteScene.IMMEDIATE_BLOCKING, this);                        // CB/SYNC
                }
                catch ( Exception ex )
                {
                    Console.WriteLine("ERROR in Prepare: {0}", ex.Message);
                }
            }
        } 

        #endregion

        #region SEND PROPERTY VALUES CLICK
        
        private void btSend_Click(object sender, System.EventArgs e)
        {
            if ( data != null )
            {
                // render-to-disk mode?
                if ( this.r2dForm == null || !this.r2dForm.RecordingCommands )
                    data.Send(ulong.MaxValue);
                else
                    data.Send(this.r2dForm.Timecode);

                RefreshExternals(this.scene2);
            }
        } 

        #endregion

        #region METHOD BUTTON CLICK
        
        private void b_Click(object sender, EventArgs e)
        {
            if ( scene2 != null )
            {
                IExternal ext = (IExternal)((Button)sender).Tag;

                // render-to-disk mode?
                if ( this.r2dForm == null || !this.r2dForm.RecordingCommands )
                    scene2.SetPropertiesInvokeMethods(null, null, new string[] { ext.Name }, null);
                else
                    scene2.SetPropertiesInvokeMethods(null, null, new string[] { ext.Name }, null, this.r2dForm.Timecode, null);
            }
        }

        private void btRefreshExt_Click(object sender, EventArgs e)
        {
            RefreshExternals(this.scene2);
        }

        #endregion

        #region REFRESH EXTERNALS
        
        private void RefreshExternals(RemoteScene rScene)
        {
            if ( rScene == null )
                return;

            // get properties and methods
            rScene.Refresh();
            IExternalCollection extColl = rScene.Externals;
            this.data = new Data2(rScene, extColl);
            this.propertyGrid1.SelectedObject = this.data;

            SetupMethodsAndEvents(extColl);
        } 

        #endregion


        #region SNAPSHOT

        /// <summary>
        /// SNAPSHOT
        /// </summary>
        private void btSnapshot_Click(object sender, System.EventArgs e)
        {
            try
            {
                if ( this.ventuz2 != null )
                {
                    System.Drawing.Image img = this.ventuz2.Snapshot();
                    if ( img != null )
                    {
                        SnashotForm snForm = new SnashotForm(img);
                        snForm.ShowDialog();
                    }
                }
            }
            catch ( Exception ex )
            {
                MessageBox.Show(this, ex.Message, "Snapshot Failed");
            }
        } 

        #endregion

        #region TIMER ELAPSED

        // TODO: is this working correctly ???
        private void timer1_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            if ( !cb_AutoUpdate.Checked )
                return;

            if ( this.ventuz2 != null )
            {
                try
                {
                    string active = null;
                    RemoteScene[] loaded = ventuz2.GetLoadedScenes();
                    IDictionaryEnumerator de = name2nodes.GetEnumerator();

                    while ( de.MoveNext() )
                    {
                        string name = ((string)de.Key).ToLower();
                        TreeNode node = (TreeNode)de.Value;

                        if ( Array.IndexOf(loaded, name) >= 0 )
                        {
                            if ( name == active )
                            {
                                if ( node.ImageIndex != 3 )
                                {
                                    node.SelectedImageIndex = 3;
                                    node.ImageIndex = 3;
                                }
                            }
                            else
                            {
                                if ( node.ImageIndex != 2 )
                                {
                                    node.SelectedImageIndex = 2;
                                    node.ImageIndex = 2;
                                }
                            }
                        }
                        else
                        {
                            if ( node.ImageIndex != 1 )
                            {
                                node.SelectedImageIndex = 1;
                                node.ImageIndex = 1;
                            }
                        }
                    }
                }
                catch
                {
                }
            }
        }  

        #endregion

        #region SELECTION CHANGES

        private void tree_AfterSelect(object sender, System.Windows.Forms.TreeViewEventArgs e)
        {
            btRender.Enabled = tree.SelectedNode != null && tree.SelectedNode.Tag != null && tree.SelectedNode.SelectedImageIndex == 1;
            btUnload.Enabled = tree.SelectedNode != null && tree.SelectedNode.SelectedImageIndex > 1;
        }

        private void propertyGrid1_SelectedObjectsChanged(object sender, System.EventArgs e)
        {
            btSend.Enabled = propertyGrid1.SelectedObject != null;
        }

        #endregion


        #region RENDER TO DISK

        Render2DiskForm r2dForm = null;

        /// <summary>
        /// Open the Render To Disk dialog
        /// </summary>
        private void bt_R2D_Click(object sender, EventArgs e)
        {
            if ( this.r2dForm == null || this.r2dForm.IsDisposed )
                this.r2dForm = new Render2DiskForm(new PlayoutConfig(), this.tb_Host.Text);
            else if ( this.r2dForm.Visible )
                return;

            this.r2dForm.FormClosed += new FormClosedEventHandler(r2dForm_FormClosed);
            this.r2dForm.Show(this);
        }

        void r2dForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            this.r2dForm.FormClosed -= new FormClosedEventHandler(r2dForm_FormClosed);
        }

        #endregion

        #region MACHINE SIGNATURE

        MachineSignatureForm machineSignForm = null;

        private void bt_MachineID_Click(object sender, EventArgs e)
        {
            if ( this.machineSignForm == null || this.machineSignForm.IsDisposed )
                this.machineSignForm = new MachineSignatureForm();
            else if ( this.machineSignForm.Visible )
                return;

            this.machineSignForm.Show(this);
        }

        #endregion

        #endregion


        #region SetButtonsAndEvents

        /// <summary>
        /// Creates Buttons to trigger extern. Methods and subscribes to extren. events!
        /// </summary>
        private void SetupMethodsAndEvents(IExternalCollection extColl)
        {
            while ( PanelRemoteMethods.Controls.Count > 0 )
            {
                PanelRemoteMethods.Controls[0].Dispose();
            }

            if ( extColl != null )
            {
                int y = 5;

                // get Event and Method names for buttons and subscribtion
                SortedDictionary<string, IExternal> bNamesDict = new SortedDictionary<string, IExternal>();
                List<string> eNamesList = new List<string>();
                IEnumerator<IExternal> extEnum = extColl.GetEnumerator();

                while ( extEnum.MoveNext() )
                {
                    IExternal ext = extEnum.Current;

                    if ( ext.MajorType == ExternalType.Method )
                    {
                        bNamesDict.Add(ext.Name, ext);
                    }
                    else if ( ext.MajorType == ExternalType.Event )
                    {
                        eNamesList.Add(ext.Name);
                    }

                    // subscribe to externalized events
                    // NOTE: all events are subscribed via scene.Callback in btRender_Click()!!!
                    if ( ext is IExternalEvent )
                    {
                        ((IExternalEvent)ext).Callback += new ExternalEventHandler(Event_Callback);
                        Console.WriteLine("-- Subscribed to Event {0}", ((IExternalEvent)ext).Name);
                    }
                }

                // created buttons in sorted order
                foreach ( KeyValuePair<string, IExternal> item in bNamesDict )
                {
                    Button b = new Button();
                    b.Location = new Point(5, y);
                    b.Size = new Size(PanelRemoteMethods.Width - 14, 20);
                    b.Text = item.Key;
                    b.Click += new EventHandler(b_Click);
                    b.Tag = item.Value;
                    b.FlatStyle = FlatStyle.System;

                    b.Anchor = AnchorStyles.Top | AnchorStyles.Left;
                    b.AutoSizeMode = AutoSizeMode.GrowAndShrink;
                    b.Dock = DockStyle.Top;

                    PanelRemoteMethods.Controls.Add(b);
                    y += 22;
                }
            }
        }

        #endregion

        #region GetNode()
        
        private TreeNode GetNode(TreeNodeCollection nodes, string[] tok, int index, string name)
        {
            if ( nodes == null || tok == null || index >= tok.Length )
                return null;

            TreeNode local = null;
            foreach ( TreeNode n in nodes )
            {
                if ( n.Text == tok[index] )
                {
                    local = n;
                    break;
                }
            }

            if ( local == null )
            {
                local = new TreeNode(tok[index]);
                nodes.Add(local);
            }

            index++;

            if ( index < tok.Length )
            {
                local.ImageIndex = 0;
                local.SelectedImageIndex = 0;
                return GetNode(local.Nodes, tok, index, name);
            }
            else
            {
                name2nodes[name] = local;
                local.SelectedImageIndex = 1;
                local.ImageIndex = 1;
                local.Tag = name;
                return local;
            }
        } 

        #endregion


        #region Event Callback Handling

        private void Scene_Callback2(RemoteScene calledScene, ulong time, string[] events, int[] args)
        {
            //Console.WriteLine("-=# SCENE_CALLBACK #=-");

            lock ( this.eventQ )
            {
                // fill Q. to process it in UpdateEvents()
                for ( int i = 0; i < events.Length; i++ )
                {
                    this.eventQ.Enqueue(string.Format("SCENE-EVENTS #{0}: {1} - Args = {2}\r\n", ++evCount, events[i], args[i]));
                    this.haveNewEvents = true;
                }
            }

            if ( this.tbEvents.InvokeRequired )
            {
                // cross-thread: so invoke on GUI-Thread
                this.tbEvents.Invoke(new UpdateEventText(this.UpdateEvents));
            }
            else
            {
                this.UpdateEvents();
            }
        }

        private void Event_Callback(IExternalEvent exernal, ulong time, int arg)
        {
            //Console.WriteLine("-=# EVENT_CALLBACK #=-");

            lock ( this.eventQ )
            {
                // fill Q. to process it in UpdateEvents()
                this.eventQ.Enqueue(string.Format("Event #{0}: {1} - Args = {2}\r\n", ++evCount, exernal.Name, arg));
                this.haveNewEvents = true;
            }

            if ( this.tbEvents.InvokeRequired )
            {
                // cross-thread: so invoke on GUI-Thread
                this.tbEvents.Invoke(new UpdateEventText(this.UpdateEvents));
            }
            else
            {
                this.UpdateEvents();
            }
        }

        private void UpdateEvents()
        {
            if ( this.haveNewEvents )
            {
                lock ( this.eventQ )
                {
                    while ( this.eventQ.Count > 0 )
                    {
                        tbEvents.Text += this.eventQ.Dequeue();
                    }

                    tbEvents.SelectionStart = tbEvents.Text.Length;
                    tbEvents.ScrollToCaret();
                    tbEvents.Update();

                    this.haveNewEvents = false;
                }
            }
        } 

        #endregion

        #region IAsyncHandler Members

        public void ASyncError(uint cmdID, ulong time, VentuzRemotingException ex)
        {
            lock ( this.eventQ )
            {
                // fill Q. to process it in UpdateEvents()
                this.eventQ.Enqueue(string.Format("Async Error: CommandID {0}  Time {1}  Error Message: {2}\r\n", cmdID, time, ex.Message));
                this.haveNewEvents = true;
            }

            if ( this.tbEvents.InvokeRequired )
            {
                // cross-thread: so invoke on GUI-Thread
                this.tbEvents.Invoke(new UpdateEventText(this.UpdateEvents));
            }
            else
            {
                this.UpdateEvents();
            }
        }

        public void SyncPoint(uint cmdID, ulong when, int tag)
        {
            lock ( this.eventQ )
            {
                // fill Q. to process it in UpdateEvents()
                this.eventQ.Enqueue(string.Format("SyncPoint: CommandID {0}  Time {1}  Tag {2}\r\n", cmdID, when, tag));
                this.haveNewEvents = true;
            }

            if ( this.tbEvents.InvokeRequired )
            {
                // cross-thread: so invoke on GUI-Thread
                this.tbEvents.Invoke(new UpdateEventText(this.UpdateEvents));
            }
            else
            {
                this.UpdateEvents();
            }
        }

        #endregion
    }
}
